@@ -6,7 +6,7 @@
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: cstdio.c,v 1.36 1999/02/16 21:13:34 guenther Exp $";
+ "$Id: cstdio.c,v 1.45 1999/11/19 07:24:43 guenther Exp $";
 #endif
 #include "procmail.h"
 #include "robust.h"
@@ -19,30 +19,62 @@ static struct dyna_long inced;				  /* includerc stack */
 struct dynstring*incnamed;
 
 void pushrc(name)const char*const name;		      /* open include rcfile */
-{ struct stat stbuf;					   /* only if size>0 */
-  stbuf.st_mode=0;
-  if(*name&&(stat(name,&stbuf)||!S_ISREG(stbuf.st_mode)||stbuf.st_size))
-   { app_val(&inced,rcbufp?(off_t)(rcbufp-rcbuf):(off_t)0);	 /* save old */
-     app_val(&inced,blasttell);app_val(&inced,(off_t)rc);   /* position & fd */
-     if(S_ISDIR(stbuf.st_mode)||bopen(name)<0)	  /* try to open the new one */
-	readerr(name),poprc();		       /* we couldn't, so restore rc */
+{ if(*name&&strcmp(name,devnull))
+   { struct stat stbuf;
+     if(stat(name,&stbuf)||!S_ISREG(stbuf.st_mode))
+	goto rerr;
+     if(stbuf.st_size)					   /* only if size>0 */
+      { app_vali(inced,rcbufp?rcbufp-rcbuf:0);			 /* save old */
+	app_valo(inced,blasttell);app_vali(inced,ifdepth);/* position, brace */
+	app_vali(inced,rc);				       /* depth & fd */
+	ifdepth=ifstack.filled;				  /* new stack depth */
+	if(bopen(name)<0)			  /* try to open the new one */
+	 { poprc();			       /* we couldn't, so restore rc */
+rerr:	   readerr(name);
+	 }
+      }
    }
 }
 
+void changerc(name)const char*const name;		    /* change rcfile */
+{ if(*name&&strcmp(name,devnull))
+   { struct stat stbuf;int orc;uchar*orbp,*orbe;
+     if(stat(name,&stbuf)||!S_ISREG(stbuf.st_mode))   /* skip irregularities */
+	goto rerr;
+     if(stbuf.st_size)		  /* only if size>0, try to open the new one */
+      { if(orbp=rcbufp,orbe=rcbufend,orc=rc,bopen(name)<0)
+	 { rcbufp=orbp;rcbufend=orbe;rc=orc;   /* we couldn't, so restore rc */
+rerr:	   readerr(name);
+	 }
+	else
+	 { struct dynstring*dp;
+	   if(dp=incnamed->enext)		      /* fixup the name list */
+	      incnamed->enext=dp->enext,free(dp);
+	   rclose(orc);
+	   ifstack.filled=ifdepth;    /* act like all the braces were closed */
+	 }
+	goto ret;
+      }
+   }
+  ifstack.filled=ifdepth;					/* see above */
+  poprc();		 /* drop the current rcfile and restore the previous */
+ret:;
+}
+
 void duprcs P((void))		/* `duplicate' all the fds of opened rcfiles */
 { size_t i;struct dynstring*dp;
   dp=incnamed;rclose(rc);
   if(0>(rc=ropen(dp->ename,O_RDONLY,0)))     /* first reopen the current one */
      goto dupfailed;
   lseek(rc,blasttell+STDBUF,SEEK_SET);	 /* you'll never know the difference */
-  for(i=inced.filled;dp=dp->enext,i;i-=2)
+  for(i=inced.filled;dp=dp->enext,i;i-=3)
    { int fd;
-     rclose(inced.offs[--i]);
+     rclose(acc_vali(inced,--i));
      if(0>(fd=ropen(dp->ename,O_RDONLY,0)))    /* reopen all (nested) others */
 dupfailed:					   /* oops, file disappeared */
 	nlog("Lost"),logqnl(dp->ename),exit(EX_NOINPUT);	    /* panic */
-     inced.offs[i]=fd; /* new & improved fd, decoupled from fd in the parent */
-   }
+     acc_vali(inced,i)=fd;		/* new & improved fd, decoupled from */
+   }							 /* fd in the parent */
 }
 
 static void closeonerc P((void))
@@ -51,30 +83,39 @@ static void closeonerc P((void))
      rclose(rc),rc= -1,last=incnamed,incnamed=last->enext,free(last);
 }
 
+static void refill(offset)const int offset;
+{ rcbufp=rcbuf+(rcbuf==(rcbufend=rcbuf+rread(rc,rcbuf,STDBUF))?1:offset);
+}
+
 int poprc P((void))
 { closeonerc();					     /* close it in any case */
-  if(skiprc)
-     skiprc=0,nlog("Missing closing brace\n");
+  if(ifstack.filled>ifdepth)		     /* force the matching of braces */
+     ifstack.filled=ifdepth,nlog("Missing closing brace\n");
+  skiprc=0;
   if(!inced.filled)				  /* include stack is empty? */
      return 0;	      /* restore rc, seekpos, prime rcbuf and restore rcbufp */
-  rc=inced.offs[--inced.filled];lseek(rc,inced.offs[--inced.filled],SEEK_SET);
-  rcbufp=rcbufend;getb();rcbufp=rcbuf+inced.offs[--inced.filled];
+  rc=acc_vali(inced,--inced.filled);
+  ifdepth=acc_vali(inced,--inced.filled);
+  blasttell=lseek(rc,acc_valo(inced,--inced.filled),SEEK_SET);
+  refill(acc_vali(inced,--inced.filled));
   return 1;
 }
 
 void closerc P((void))					/* {while(poprc());} */
 { while(closeonerc(),inced.filled)
-     rc=inced.offs[inced.filled-1],inced.filled-=3;
+     rc=acc_vali(inced,inced.filled-1),inced.filled-=4;
+  ifstack.filled=ifdepth=0;
 }
 							    /* destroys buf2 */
 int bopen(name)const char*const name;				 /* my fopen */
 { rcbufp=rcbufend=0;rc=ropen(name,O_RDONLY,0);
   if(rc>=0)
-   { char*md;		 /* if it's a relative name and an absolute $MAILDIR */
+   { char*md;size_t len; /* if it's a relative name and an absolute $MAILDIR */
      if(!strchr(dirsep,*name)&&
 	*(md=(char*)tgetenv(maildir))&&
-	strchr(dirsep,*md))
-      { strcpy(buf2,md);*(md=strchr(buf2,'\0'))= *dirsep;strcpy(++md,name);
+	strchr(dirsep,*md)&&
+	(len=strlen(md))+strlen(name)+2+XTRAlinebuf<linebuf)
+      { strcpy(buf2,md);*(md=buf2+len)= *dirsep;strcpy(++md,name);
 	md=buf2;				    /* then prepend $MAILDIR */
       }
      else
@@ -99,10 +140,7 @@ int getbl(p,end)char*p,*end;					  /* my gets */
 
 int getb P((void))						 /* my fgetc */
 { if(rcbufp==rcbufend)						   /* refill */
-   { blasttell=tell(rc);rcbufend=rcbuf+rread(rc,rcbufp=rcbuf,STDBUF);
-     if(rcbufp==rcbufend)
-	rcbufp++;
-   }
+     blasttell=tell(rc),refill(0);
   return rcbufp<rcbufend?(int)*rcbufp++:EOF;
 }
 
@@ -140,9 +178,9 @@ void skipline P((void))
       }
 }
 
-int getlline(target)char*target;
-{ char*chp2,*end;int overflow;
-  for(end=target+linebuf,overflow=0;;*target++='\n')
+int getlline(target,end)char*target,*end;
+{ char*chp2;int overflow;
+  for(overflow=0;;*target++='\n')
      switch(getbl(chp2=target,end))			   /* read line-wise */
       { case -1:overflow=1;
 	case 1:
