@@ -3,14 +3,14 @@
  *									*
  *	Seems to be relatively bug free.				*
  *									*
- *	Copyright (c) 1990-1994, S.R. van den Berg, The Netherlands	*
+ *	Copyright (c) 1990-1999, S.R. van den Berg, The Netherlands	*
  *	#include "../README"						*
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: formail.c,v 1.73 1994/10/31 17:30:23 berg Exp $";
+ "$Id: formail.c,v 1.97 1999/10/24 06:32:08 guenther Exp $";
 #endif
-static /*const*/char rcsdate[]="$Date: 1994/10/31 17:30:23 $";
+static /*const*/char rcsdate[]="$Date: 1999/10/24 06:32:08 $";
 #include "includes.h"
 #include <ctype.h>		/* iscntrl() */
 #include "formail.h"
@@ -21,6 +21,7 @@ static /*const*/char rcsdate[]="$Date: 1994/10/31 17:30:23 $";
 #include "fields.h"
 #include "ecommon.h"
 #include "formisc.h"
+#include "../patchlevel.h"
 
 #define ssl(str)		str,STRLEN(str)
 #define bsl(str)		{ssl(str)}
@@ -35,7 +36,8 @@ static const char
  x_[]=			"X-",				/* general extension */
  old_[]=		OLD_PREFIX,			     /* my extension */
  xloop[]=		"X-Loop:",				/* ditto ... */
- daemon[]="<>",unknown[]=UNKNOWN,re[]=" Re:",fmusage[]=FM_USAGE;
+ Resent_[]=		"Resent-",	   /* for tweaking reply preferences */
+ mdaemon[]="<>",unknown[]=UNKNOWN,re[]=" Re:",fmusage[]=FM_USAGE;
 
 static const struct {const char*hedr;int lnr;}cdigest[]=
 {
@@ -47,22 +49,26 @@ static const struct {const char*hedr;int lnr;}cdigest[]=
 /*
  *	sender determination fields in order of importance/reliability
  *	reply-address determination fields (wrepl specifies the weight
- *	for regular replies, wtrepl specifies the weight for trusted users)
+ *	for header replies and wrrepl specifies the weight for header
+ *	replies where Resent- header are used, while the position in the
+ *	table index specifies the weight for envelope replies and From_
+ *	line creation.
  *
- *	I bet this is the first time you see a bar graph in C-source-code :-)
+ *	I bet this is the first time you've seen a bar graph in
+ *	C-source-code :-)
  */
-static const struct {const char*head;int len,wrepl,wtrepl;}sest[]=
-{ sslbar(replyto	,"******"	,"********"	),
-  sslbar(Fromm		,"*"		,"*******"	),
-  sslbar(retreceiptto	,"********"	,"*****"	),
-  sslbar(sender		,"*****"	,"******"	),
-  sslbar(res_replyto	,"***********"	,"***********"	),
-  sslbar(res_from	,"***foo***"	,"***bar****"	),
-  sslbar(res_sender	,"**********"	,"*********"	),
-  sslbar(errorsto	,"*******"	,"****"		),
+static const struct {const char*head;int len,wrepl,wrrepl;}sest[]=
+{ sslbar(replyto	,"*********"	,"********"	),
+  sslbar(Fromm		,"**foo***"	,"**bar**"	),
+  sslbar(sender		,"*******"	,"******"	),
+  sslbar(res_replyto	,"*"		,"***********"	),
+  sslbar(res_from	,"*"		,"**********"	),
+  sslbar(res_sender	,"*"		,"*********"	),
   sslbar(path		,"**"		,"*"		),
-  sslbar(returnpath	,"***"		,"***"		),
-  sslbar(From_		,"****"		,"**"		)
+  sslbar(retreceiptto	,"***"		,"**"		),
+  sslbar(errorsto	,"****"		,"***"		),
+  sslbar(returnpath	,"******"	,"*****"	),
+  sslbar(From_		,"*****"	,"****"		),
 };
 
 static struct saved rex[]=
@@ -88,38 +94,39 @@ static const char emboxsep[]=eMAILBOX_SEPARATOR;
 #endif /* eMAILBOX_SEPARATOR */
 
 const char binsh[]=BinSh,sfolder[]=FOLDER,
- couldntw[]="Couldn't write to stdout";
+ couldntw[]="Couldn't write to stdout",formailn[]=FORMAILN;
 int errout,oldstdout,quiet=1,zap,buflast,lenfileno;
 long initfileno;
 char ffileno[LEN_FILENO_VAR+8*sizeof(initfileno)*4/10+1+1]=DEFfileno;
 int lexitcode;					     /* dummy, for waitfor() */
 pid_t child= -1;
+int childlimit;
 unsigned long rhash;
 FILE*mystdout;
 int nrskip,nrtotal= -1,retval=EXIT_SUCCESS;
 size_t buflen,buffilled;
-long totallen;
+long Totallen;
 char*buf,*logsummary;
 struct field*rdheader,*xheader,*Xheader,*uheader,*Uheader;
 static struct field*iheader,*Iheader,*aheader,*Aheader,*Rheader,*nheader;
 static int areply;
 
 static void logfolder P((void))	 /* estimate the no. of characters needed to */
-{ size_t i;charNUM(num,totallen);		       /* represent totallen */
+{ size_t i;charNUM(num,Totallen);		       /* represent Totallen */
   static const char tabchar[]=TABCHAR;
   if(logsummary)
    { putssn(sfolder,STRLEN(sfolder));putssn(logsummary,i=strlen(logsummary));
      i+=STRLEN(sfolder);i-=i%TABWIDTH;
      do putssn(tabchar,STRLEN(tabchar));
      while((i+=TABWIDTH)<LENoffset);
-     ultstr(7,totallen,num);putssn(num,strlen(num));putcs('\n');
+     ultstr(7,Totallen,num);putssn(num,strlen(num));putcs('\n');
    }
 }
 
 static void renfield(pointer,oldl,newname,newl)struct field**const pointer;
  size_t oldl;const size_t newl;const char*const newname;    /* rename fields */
 { struct field*p;size_t i;char*chp;
-  p= *pointer;(chp=p->fld_text)[p->tot_len-1]='\0';
+  p= *pointer;(chp=p->fld_text)[p->Tot_len-1]='\0';
   if(eqFrom_(chp))				       /* continued From_ to */
      for(;chp=strstr(chp,"\n>");*++chp=' ');	  /* continued regular field */
   if(newl==STRLEN(From_)&&eqFrom_(newname))
@@ -132,10 +139,10 @@ static void renfield(pointer,oldl,newname,newl)struct field**const pointer;
   if(newname[newl-1]==HEAD_DELIMITER)		     /* completely new field */
 replaceall:
      oldl=p->id_len;			     /* replace the old one entirely */
-  p->id_len+=(int)newl-(int)oldl;p->fld_text[p->tot_len-1]='\n';
-  p->tot_len=(i=p->tot_len-oldl)+newl;
+  p->id_len+=(int)newl-(int)oldl;p->fld_text[p->Tot_len-1]='\n';
+  p->Tot_len=(i=p->Tot_len-oldl)+newl;
   if(newl>oldl)
-     *pointer=p=realloc(p,FLD_HEADSIZ+p->tot_len);
+     *pointer=p=realloc(p,FLD_HEADSIZ+p->Tot_len);
   chp=p->fld_text;tmemmove(chp+newl,chp+oldl,i);tmemmove(chp,newname,newl);
 }
 
@@ -144,28 +151,32 @@ static void procfields(sareply)const int sareply;
   fldp= *(afldp= &rdheader);
   while(fldp)
    { struct field*fp2;
-     if(fp2=findf(fldp,&Rheader))		  /* explicitly rename field */
-      { renfield(afldp,fp2->id_len,(char*)fp2->fld_text+fp2->id_len,
-	 fp2->tot_len-fp2->id_len);
-	goto fixfldp;
-      }
      if(!sareply&&
 	(fp2=findf(fldp,&iheader))&&
-	!(areply&&fldp->id_len>=fp2->tot_len-1))      /* filled replacement? */
+	!(areply&&fldp->id_len>=fp2->Tot_len-1))      /* filled replacement? */
       { renfield(afldp,(size_t)0,old_,STRLEN(old_));	/* implicitly rename */
+	goto fixfldp;
+      }
+     if((fp2=findf(fldp,&Iheader))&&			    /* delete fields */
+	!(sareply&&fldp->id_len<fp2->Tot_len-1))       /* empty replacement? */
+	goto delfld;
+     if(fp2=findf(fldp,&Rheader))		  /* explicitly rename field */
+      { renfield(afldp,fp2->id_len,(char*)fp2->fld_text+fp2->id_len,
+	 fp2->Tot_len-fp2->id_len);
 fixfldp:
 	fldp= *afldp;
       }
-     if((fp2=findf(fldp,&Iheader))&&			    /* delete fields */
-	!(sareply&&fldp->id_len<fp2->tot_len-1))       /* empty replacement? */
-	goto delfld;
      ;{ struct field*uf;
 	if((uf=findf(fldp,&uheader))&&!uf->fld_ref)
 	   uf->fld_ref=afldp;			   /* first uheader, keep it */
 	else if(fp2=findf(fldp,&Uheader))
 	 { if(fp2->fld_ref)
-	    { if(afldp==&(*fp2->fld_ref)->fld_next)
-		 afldp=fp2->fld_ref;
+	    { struct field**ch_afldp;
+	      if(afldp==(ch_afldp= &(*fp2->fld_ref)->fld_next))
+		 afldp=fp2->fld_ref;		   /* deleting own reference */
+	      for(fldp=Uheader;fldp;fldp=fldp->fld_next)
+		 if(fldp->fld_ref==ch_afldp)	  /* rearrange references to */
+		    fldp->fld_ref=fp2->fld_ref;		  /* vanishing field */
 	      delfield(fp2->fld_ref);		       /* delete old Uheader */
 	    }
 	   fp2->fld_ref=afldp;				/* keep last Uheader */
@@ -196,19 +207,19 @@ static int artheadr P((void))	     /* could it be the start of an article? */
   return 0;
 }
 			     /* lifted out of main() to reduce main()'s size */
-static char*getsender(namep,fldp,trust)char*namep;struct field*fldp;
- const int trust;
+static char*getsender(namep,fldp,headreply)char*namep;struct field*fldp;
+ const int headreply;
 { char*chp;int i,nowm;size_t j;static int lastm;
   chp=fldp->fld_text;j=fldp->id_len;i=maxindex(sest);
   while((sest[i].len!=j||strnIcmp(sest[i].head,chp,j))&&i--);
   if(i>=0&&(i!=maxindex(sest)||fldp==rdheader))		  /* found anything? */
    { char*saddr;char*tmp;			     /* determine the weight */
-     nowm=trust?sest[i].wtrepl:areply?sest[i].wrepl:i;chp+=j;
-     tmp=malloc(j=fldp->tot_len-j);tmemmove(tmp,chp,j);(chp=tmp)[j-1]='\0';
+     nowm=areply&&headreply?headreply==1?sest[i].wrepl:sest[i].wrrepl:i;chp+=j;
+     tmp=malloc(j=fldp->Tot_len-j);tmemmove(tmp,chp,j);(chp=tmp)[j-1]='\0';
      if(sest[i].head==From_)
       { char*pastad;
-	if(trust||!(saddr=strchr(chp,'\n')))	     /* skip the first line? */
-	   saddr=chp;						  /* no need */
+	if(strchr(saddr=chp,'\n'))		     /* multiple From_ lines */
+	   nowm-=2;				    /* aren't as trustworthy */
 	if(*saddr=='\n'&&(pastad=strchr(saddr,' ')))
 	   saddr=pastad+1;			/* reposition at the address */
 	chp=saddr;
@@ -216,8 +227,8 @@ static char*getsender(namep,fldp,trust)char*namep;struct field*fldp;
 	   chp=pastad+1;		      /* skip to the last uucp >From */
 	if(pastad=strchr(chp,' '))			/* found an address? */
 	 { char*savetmp;				      /* lift it out */
-	   savetmp=malloc((j=pastad-chp)+1);tmemmove(savetmp,chp,j);
-	   savetmp[j]='\0';
+	   savetmp=malloc(1+(j=pastad-chp)+1+1);tmemmove(savetmp,chp,j);
+	   savetmp[j]='\0';				   /* make work copy */
 	   if(strchr(savetmp,'@'))			 /* domain attached? */
 	      chp=savetmp,savetmp=tmp,tmp=chp;			/* ok, ready */
 	   else					/* no domain, bang away! :-) */
@@ -244,8 +255,9 @@ static char*getsender(namep,fldp,trust)char*namep;struct field*fldp;
 		 saddr=p1;				/* continue the hunt */
 	       }
 	      strcpy(chp,savetmp);chp=tmp;	     /* attach the user part */
-	    }
-	   free(savetmp);	  /* (temporary buffers might have switched) */
+	    }			  /* (temporary buffers might have switched) */
+	   free(savetmp);savetmp=strchr(tmp,'\0');	      /* prepend '<' */
+	   tmemmove(tmp+1,tmp,savetmp-tmp);*tmp='<';savetmp[1]='\0';
 	 }
       }
      while(*(chp=skpspace(chp))=='\n')
@@ -275,7 +287,7 @@ static char*getsender(namep,fldp,trust)char*namep;struct field*fldp;
 	      goto pnewname;
 	 }
 	else if(sest[i].head==returnpath)		/* nill Return-Path: */
-	 { saddr=(char*)daemon;nowm=maxindex(sest)+2;		 /* override */
+	 { saddr=(char*)mdaemon;nowm=maxindex(sest)+2;		 /* override */
 pnewname:  lastm=nowm;saddr=strcpy(malloc(strlen(saddr)+1),saddr);
 	   if(namep)
 	      free(namep);
@@ -286,14 +298,71 @@ pnewname:  lastm=nowm;saddr=strcpy(malloc(strlen(saddr)+1),saddr);
    }					   /* save headers for later perusal */
   return namep;
 }
+			     /* lifted out of main() to reduce main()'s size */
+static void elimdups(namep,idcache,maxlen,split)const char*const namep;
+ FILE*idcache;const off_t maxlen;const int split;
+{ int dupid=0;char*key,*oldnewl;
+  key=(char*)namep;		  /* not to worry, no change will be noticed */
+  if(!areply)
+   { key=0;
+     if(msid->rexl)					/* any Message-ID: ? */
+	*(oldnewl=(key=msid->rexp)+msid->rexl-1)='\0';
+   }						/* wipe out trailing newline */
+  if(key)
+   { off_t insoffs=maxlen;
+     while(*key==' ')				     /* strip leading spaces */
+	key++;
+     do
+      { int j;char*p;		  /* start reading & comparing the next word */
+	for(p=key;(j=fgetc(idcache))==*p;p++)
+	   if(!j)					     /* end of word? */
+	    { if(!quiet)
+		 nlog("Duplicate key found:"),elog(key),elog("\n");
+	      dupid=1;
+	      goto dupfound;			     /* YES! duplicate found */
+	    }
+	if(!j)						     /* end of word? */
+	 { if(p==key&&insoffs==maxlen)			 /* first character? */
+	    { insoffs=ftell(idcache)-1;			     /* found end of */
+	      goto skiprest;				  /* circular buffer */
+	    }
+	 }
+	else
+skiprest:  for(;;)				/* skip the rest of the word */
+	    { switch(fgetc(idcache))
+	       { case EOF:
+		    goto noluck;
+		 default:
+		    continue;
+		 case '\0':;
+	       }
+	      break;
+	    }
+      }
+     while(ftell(idcache)<maxlen);			  /* past our quota? */
+noluck:
+     if(insoffs>=maxlen)				  /* past our quota? */
+	insoffs=0;				     /* start up front again */
+     fseek(idcache,insoffs,SEEK_SET);fwrite(key,1,strlen(key)+1,idcache);
+     putc('\0',idcache);			   /* mark new end of buffer */
+dupfound:
+     fseek(idcache,(off_t)0,SEEK_SET);		 /* rewind, for any next run */
+     if(!areply)
+	*oldnewl='\n';				      /* restore the newline */
+   }
+  if(!split)				  /* not splitting?  terminate early */
+     exit(dupid?EXIT_SUCCESS:1);
+  if(dupid)				       /* duplicate? suppress output */
+     closemine(),opensink();
+}
 
 static PROGID;
 
-main(lastm,argv)int lastm;const char*const argv[];
-{ int i,split=0,force=0,bogus=1,every=0,trust=0,digest=0,nowait=0,keepb=0,
+int main(lastm,argv)int lastm;const char*const argv[];
+{ int i,split=0,force=0,bogus=1,every=0,headreply=0,digest=0,nowait=0,keepb=0,
    minfields=(char*)progid-(char*)progid,conctenate=0,babyl=0,babylstart,
    berkeley=0,forgetclen;
-  off_t maxlen,insoffs,ctlength;FILE*idcache=0;pid_t thepid;
+  off_t maxlen,ctlength;FILE*idcache=0;pid_t thepid;
   size_t j,lnl,escaplen;char*chp,*namep,*escap=ESCAP;
   struct field*fldp,*fp2,**afldp,*fdate,*fcntlength,*fsubject,*fFrom_;
   if(lastm)			       /* sanity check, any argument at all? */
@@ -305,7 +374,7 @@ main(lastm,argv)int lastm;const char*const argv[];
 	   goto usg;
 	for(;;)
 	 { switch(lastm= *chp++)
-	    { case FM_TRUST:trust=1;
+	    { case FM_TRUST:headreply|=1;
 		 continue;
 	      case FM_REPLY:areply=1;
 		 continue;
@@ -316,7 +385,8 @@ main(lastm,argv)int lastm;const char*const argv[];
 	      case FM_BABYL:babyl=every=1;
 	      case FM_DIGEST:digest=1;
 		 continue;
-	      case FM_NOWAIT:nowait=1;
+	      case FM_NOWAIT:nowait=1;Qnext_arg();
+		 childlimit=strtol(chp,&chp,10);
 		 continue;
 	      case FM_KEEPB:keepb=1;
 		 continue;
@@ -353,7 +423,7 @@ number:		 if(*chp-'0'>(unsigned)9)	    /* the number is not >=0 */
 		    case FM_DUPLICATE:maxlen=i;Qnext_arg();
 		       if(!(idcache=fopen(chp,"r+b"))&&	  /* existing cache? */
 			  !(idcache=fopen(chp,"w+b")))	    /* create cache? */
-			{ nlog("Couldn't open");logqnl(argv[i]);
+			{ nlog("Couldn't open");logqnl(chp);
 			  return EX_CANTCREAT;
 			}
 		       goto nextarg;
@@ -368,14 +438,24 @@ number:		 if(*chp-'0'>(unsigned)9)	    /* the number is not >=0 */
 		 continue;
 	      case FM_QPREFIX:Qnext_arg();escap=chp;
 		 break;
+	      case FM_VERSION:elog(formailn);elog(VERSION);
+		 goto xusg;
 	      case FM_ADD_IFNOT:case FM_ADD_ALWAYS:case FM_REN_INSERT:
 	      case FM_DEL_INSERT:case FM_EXTRACT:case FM_EXTRC_KEEP:
 	      case FM_FIRST_UNIQ:case FM_LAST_UNIQ:case FM_ReNAME:Qnext_arg();
 		 i=breakfield(chp,lnl=strlen(chp));
 		 switch(lastm)
-		  { default:
+		  { case FM_ADD_IFNOT:
+		       if(i>0)			   /* the only partial field */
+			  break;			  /* allowed with -a */
+		       else if(i!=-STRLEN(Resent_)||-i!=lnl||  /* is Resent- */
+			strnIcmp(chp,Resent_,STRLEN(Resent_)+1))
+			  goto invfield;
+		       headreply|=2;
+		       goto nextarg;		/* don't add to the list */
+		    default:
 		       if(-i!=lnl)	  /* it is not an early ending field */
-		    case FM_ADD_IFNOT:case FM_ADD_ALWAYS:
+		    case FM_ADD_ALWAYS:
 			  if(i<=0)	      /* and it is not a valid field */
 			     goto invfield;			 /* complain */
 		    case FM_ReNAME:;		       /* everything allowed */
@@ -406,13 +486,14 @@ number:		 if(*chp-'0'>(unsigned)9)	    /* the number is not >=0 */
 		       tmemmove((char*)fldp->fld_text+lnl,chp,i),copied=1;
 		    else if(namep>chp||				 /* garbage? */
 			    !(chp=(char*)*++argv)||	 /* look at next arg */
-			    !(i=breakfield(chp,strlen(chp)))||	/* fieldish? */
-			    i<0&&(i= -i,lastm>0))  /* impossible combination */
+			    (!(i=breakfield(chp,strlen(chp)))&& /* fieldish? */
+			     *chp)||			   /* but "" is fine */
+			    i<=0&&(i= -i,lastm>0)) /* impossible combination */
 invfield:	     { nlog("Invalid field-name:");logqnl(chp?chp:"");
 		       goto usg;
 		     }
 		    *afldp=fldp=
-		     realloc(fldp,FLD_HEADSIZ+(fldp->tot_len=lnl+i));
+		     realloc(fldp,FLD_HEADSIZ+(fldp->Tot_len=lnl+i));
 		    if(!copied)			   /* if not squeezed on yet */
 		       tmemmove((char*)fldp->fld_text+lnl,chp,i);  /* do now */
 		  }
@@ -424,9 +505,25 @@ nextarg:;
       }
 parsedoptions:
   escaplen=strlen(escap);mystdout=stdout;signal(SIGPIPE,SIG_IGN);
+  if(nowait&&idcache&&split)
+   { nowait=childlimit=0;
+     if(!quiet)
+	nlog("No-wait option ignored to prevent corrupted idcache\n");
+   }
+#ifdef SIGCHLD
+  signal(SIGCHLD,SIG_DFL);
+#endif
   thepid=getpid();
+  if(babyl)						/* skip BABYL leader */
+   { while(getchar()!=BABYL_SEP1||getchar()!=BABYL_SEP2||getchar()!='\n')
+	while(getchar()!='\n');
+     while(getchar()!='\n');
+   }
+  while((buflast=getchar())=='\n');		     /* skip leading garbage */
   if(split)
    { char**ep;char**vfileno=0;
+     if(buflast==EOF)			   /* avoid splitting empty messages */
+	return EXIT_SUCCESS;
      for(ep=environ;*ep;ep++)		   /* gobble through the environment */
 	if(!strncmp(*ep,ffileno,LEN_FILENO_VAR))	 /* look for FILENO= */
 	   vfileno=ep;					    /* yes, found it */
@@ -460,7 +557,11 @@ usg:						     /* options sanity check */
 xusg:
      return EX_USAGE;
    }
-  buf=malloc(buflen=Bsize);totallen=0;i=maxindex(rex); /* prime some buffers */
+  if(headreply==2)				/* -aResent- is only allowed */
+   { chp=(char*)Resent_;		  /* as a modifier to header replies */
+     goto invfield;
+   }
+  buf=malloc(buflen=Bsize);Totallen=0;i=maxindex(rex); /* prime some buffers */
   do rex[i].rexp=malloc(1);
   while(i--);
   fdate=0;addfield(&fdate,date,STRLEN(date)); /* fdate is only for searching */
@@ -475,12 +576,6 @@ xusg:
 	       !findf(fcntlength,&Xheader));	  /* getting Content-Length: */
   if(areply)					       /* when auto-replying */
      addfield(&iheader,xloop,STRLEN(xloop));	  /* preserve X-Loop: fields */
-  if(babyl)						/* skip BABYL leader */
-   { while(getchar()!=BABYL_SEP1||getchar()!=BABYL_SEP2||getchar()!='\n')
-	while(getchar()!='\n');
-     while(getchar()!='\n');
-   }
-  while((buflast=getchar())=='\n');		     /* skip leading garbage */
   if(!readhead())					    /* start looking */
    {
 #ifdef sMAILBOX_SEPARATOR			      /* check for a leading */
@@ -505,10 +600,10 @@ startover:
 		!strncmp(tmp,mailfrom,STRLEN(mailfrom))&&
 		eqFrom_(tmp2=skpspace(tmp+STRLEN(mailfrom))))
 	 { rdheader->id_len=STRLEN(From_);
-	   tmemmove(tmp,tmp2,rdheader->tot_len-=tmp2-tmp);
+	   tmemmove(tmp,tmp2,rdheader->Tot_len-=tmp2-tmp);
 	 }
       }
-     namep=0;totallen=0;i=maxindex(rex);
+     namep=0;Totallen=0;i=maxindex(rex);
      do rex[i].rexl=0;
      while(i--);			      /* reset all state information */
      clear_uhead(uheader);clear_uhead(Uheader);
@@ -518,82 +613,33 @@ startover:
       { if(zap)		      /* go through the linked list of header-fields */
 	 { chp=fldp->fld_text+(j=fldp->id_len);
 	   if(chp[-1]==HEAD_DELIMITER)
-	      if(*chp!=' '&&fldp->tot_len>j+1)
+	      if((*chp!=' '&&*chp!='\t')&&fldp->Tot_len>j+1)
 	       { chp=j+(*afldp=fldp=
-		  realloc(fldp,FLD_HEADSIZ+(i= ++fldp->tot_len)))->fld_text;
+		  realloc(fldp,FLD_HEADSIZ+(i=fldp->Tot_len++)+1))->fld_text;
 		 tmemmove(chp+1,chp,i-j);*chp=' ';
 	       }
-	      else if(fldp->tot_len<=j+2)
+	      else if(fldp->Tot_len<=j+2)
 	       { *afldp=fldp->fld_next;free(fldp);fldp= *afldp;
 		 continue;
 	       }
 	 }
 	if(conctenate)
 	   concatenate(fldp);		    /* save fields for later perusal */
-	namep=getsender(namep,fldp,trust);i=maxindex(rex);chp=fldp->fld_text;
-	j=fldp->id_len;
+	namep=getsender(namep,fldp,headreply);
+	i=maxindex(rex);chp=fldp->fld_text;j=fldp->id_len;
 	while((rex[i].lenr!=j||strnIcmp(rex[i].headr,chp,j))&&i--);
 	chp+=j;
-	if(i>=0&&(j=fldp->tot_len-j)>1)			  /* found anything? */
+	if(i>=0&&(j=fldp->Tot_len-j)>1)			  /* found anything? */
 	 { tmemmove(rex[i].rexp=realloc(rex[i].rexp,(rex[i].rexl=j)+1),chp,j);
 	   rex[i].rexp[j]='\0';			     /* add a terminating \0 */
 	 }
 	fldp= *(afldp= &fldp->fld_next);
       }
      if(idcache)
-      { int dupid=0;char*key;
-	key=namep;
-	if(!areply)
-	 { key=0;
-	   if(msid->rexl)				/* any Message-ID: ? */
-	      (key=msid->rexp)[msid->rexl-1]='\0';
-	 }
-	if(key)
-	 { insoffs=maxlen;
-	   do					/* wipe out trailing newline */
-	    { int j;char*p;	  /* start reading & comparing the next word */
-	      for(p=key;(j=fgetc(idcache))==*p;p++)
-		 if(!j)					     /* end of word? */
-		  { if(!quiet)
-		       nlog("Duplicate key found:"),elog(key),elog("\n");
-		    dupid=1;
-		    goto dupfound;		     /* YES! duplicate found */
-		  }
-	      if(!j)					     /* end of word? */
-	       { if(p==key&&insoffs==maxlen)		 /* first character? */
-		  { insoffs=ftell(idcache)-1;		     /* found end of */
-		    goto skiprest;			  /* circular buffer */
-		  }
-	       }
-	      else
-skiprest:	 for(;;)			/* skip the rest of the word */
-		  { switch(fgetc(idcache))
-		     { case EOF:
-			  goto noluck;
-		       default:
-			  continue;
-		       case '\0':;
-		     }
-		    break;
-		  }
-	    }
-	   while(ftell(idcache)<maxlen);		  /* past our quota? */
-noluck:	   if(insoffs>=maxlen)				  /* past our quota? */
-	      insoffs=0;			     /* start up front again */
-	   fseek(idcache,insoffs,SEEK_SET);fwrite(key,1,strlen(key)+1,idcache);
-	   putc('\0',idcache);			   /* mark new end of buffer */
-dupfound:  fseek(idcache,(off_t)0,SEEK_SET);	 /* rewind, for any next run */
-	   if(!areply)
-	      key[msid->rexl-1]='\n';		      /* restore the newline */
-	 }
-	if(!split)			  /* not splitting?  terminate early */
-	   return dupid?EXIT_SUCCESS:1;
-	if(dupid)			       /* duplicate? suppress output */
-	   closemine(),opensink();
-      }
+	elimdups(namep,idcache,maxlen,split);
      ctlength=0;
      if(!forgetclen&&(fldp=findf(fcntlength,&rdheader)))
-      { *(chp=(char*)fldp->fld_text+fldp->tot_len-1)='\0';   /* terminate it */
+      { *(chp=(char*)fldp->fld_text+fldp->Tot_len-1)='\0';   /* terminate it */
 	ctlength=strtol((char*)fldp->fld_text+STRLEN(cntlength),(char**)0,10);
 	*chp='\n';			     /* restore the trailing newline */
       }
@@ -601,15 +647,15 @@ dupfound:  fseek(idcache,(off_t)0,SEEK_SET);	 /* rewind, for any next run */
      buffilled=0;    /* moved the contents of buf out of the way temporarily */
      if(areply)		      /* autoreply requested, we clean up the header */
       { for(fldp= *(afldp= &rdheader);fldp;)
-	   if(!(fp2=findf(fldp,&iheader))||fp2->id_len<fp2->tot_len-1)
+	   if(!(fp2=findf(fldp,&iheader))||fp2->id_len<fp2->Tot_len-1)
 	      *afldp=fldp->fld_next,free(fldp),fldp= *afldp;   /* remove all */
 	   else					/* except the ones mentioned */
 	      fldp= *(afldp= &fldp->fld_next);		       /* as -i ...: */
-	loadbuf(to,STRLEN(to));loadchar(' ');	   /* generate the To: field */
+	loadbuf(To,STRLEN(To));loadchar(' ');	   /* generate the To: field */
 	if(namep)	       /* did we find a valid return address at all? */
 	   loadbuf(namep,strlen(namep));	      /* then insert it here */
-	else
-	   loadbuf(unknown,STRLEN(unknown));	    /* or insert our default */
+	else					    /* or insert our default */
+	   retval=EX_NOUSER,loadbuf(unknown,STRLEN(unknown));
 	loadchar('\n');addbuf();		       /* add it to rdheader */
 	if(subj->rexl)				      /* any Subject: found? */
 	 { loadbuf(subject,STRLEN(subject));	  /* sure, check for leading */
@@ -634,7 +680,7 @@ dupfound:  fseek(idcache,(off_t)0,SEEK_SET);	 /* rewind, for any next run */
       }
      else if(!force&&		       /* are we allowed to add From_ lines? */
 	     (!rdheader||!eqFrom_(rdheader->fld_text))&&   /* is it missing? */
-	     ((fldp=findf(fFrom_,&aheader))&&STRLEN(From_)+1>=fldp->tot_len||
+	     ((fldp=findf(fFrom_,&aheader))&&STRLEN(From_)+1>=fldp->Tot_len||
 	      !wasafrom_&&			    /* if there was no From_ */
 	      !findf(fFrom_,&iheader)&&		   /* and From_ is not being */
 	      !findf(fFrom_,&Iheader)&&				/* supressed */
@@ -655,7 +701,7 @@ dupfound:  fseek(idcache,(off_t)0,SEEK_SET);	 /* rewind, for any next run */
       }
      for(fldp=aheader;fldp;fldp=fldp->fld_next)
 	if(!findf(fldp,&rdheader))	       /* only add what didn't exist */
-	   if(fldp->id_len+1>=fldp->tot_len&&		  /* field name only */
+	   if(fldp->id_len+1>=fldp->Tot_len&&		  /* field name only */
 	      (fldp->id_len==STRLEN(messageid)&&
 	       !strnIcmp(fldp->fld_text,messageid,STRLEN(messageid))||
 	       fldp->id_len==STRLEN(res_messageid)&&
@@ -663,22 +709,21 @@ dupfound:  fseek(idcache,(off_t)0,SEEK_SET);	 /* rewind, for any next run */
 	    { char*p;const char*name;unsigned long h1,h2,h3;
 	      static unsigned long h4; /* conjure up a `unique' msg-id field */
 	      h1=time((time_t*)0);h2=thepid;h3=rhash;
-	      p=chp=malloc(fldp->id_len+2+1+((sizeof h1*8+5)/6+1)*4+1+
+	      p=chp=malloc(fldp->id_len+2+((sizeof h1*8+5)/6+1)*4+
 	       strlen(name=hostname())+2);     /* allocate worst case length */
 	      strncpy(p,fldp->fld_text,fldp->id_len);*(p+=fldp->id_len)=' ';
-	      *++p='<';*++p='"';*(p=ultoan(h3,p+1))='.';
-	      *(p=ultoan(h4,p+1))='.';*(p=ultoan(h2,p+1))='.';
-	      *(p=ultoan(h1,p+1))='"';*++p='@';strcpy(p+1,name);
+	      *++p='<';*(p=ultoan(h3,p+1))='.';*(p=ultoan(h4,p+1))='.';
+	      *(p=ultoan(h2,p+1))='.';*(p=ultoan(h1,p+1))='@';strcpy(p+1,name);
 	      *(p=strchr(p,'\0'))='>';*++p='\n';addfield(&nheader,chp,p-chp+1);
 	      free(chp);h4++;					/* put it in */
 	    }
 	   else
-	      addfield(&nheader,fldp->fld_text,fldp->tot_len);
+	      addfield(&nheader,fldp->fld_text,fldp->Tot_len);
      if(logsummary)
       { if(eqFrom_(rdheader->fld_text))
-	   putssn(rdheader->fld_text,rdheader->tot_len);
+	   putssn(rdheader->fld_text,rdheader->Tot_len);
 	if(fldp=findf(fsubject,&rdheader))
-	 { concatenate(fldp);(chp=fldp->fld_text)[i=fldp->tot_len-1]='\0';
+	 { concatenate(fldp);(chp=fldp->fld_text)[i=fldp->Tot_len-1]='\0';
 	   detab(chp);putcs(' ');
 	   putssn(chp,i>=MAXSUBJECTSHOW?MAXSUBJECTSHOW:i);putcs('\n');
 	 }
@@ -751,11 +796,11 @@ splitit:       { if(!lnl)   /* did the previous mail end with an empty line? */
 		 if(!nowait&&*argv)	 /* wait till the child has finished */
 		  { int excode;
 		    if((excode=waitfor(child))!=EXIT_SUCCESS&&
-		       retval!=EXIT_SUCCESS)
+		       retval==EXIT_SUCCESS)
 		       retval=excode;
 		  }
 		 if(!nrtotal)
-		    goto onlyhead;
+		    goto nconlyhead;
 		 startprog((const char*Const*)argv);
 		 goto startover;
 	       }				    /* and there we go again */
@@ -779,10 +824,14 @@ splitit:       { if(!lnl)   /* did the previous mail end with an empty line? */
 	      goto splitit;
 	   break;
 	 }
-	else if(bogus)
+#ifdef eMAILBOX_SEPARATOR
+	if(buflast==EOF)
+	   break;
+#endif
+	if(bogus)
 	   goto putsp;				   /* escape it with a space */
       }
-     else if(!strncmp(smboxsep,buf,STRLEN(smboxsep)&&bogus))
+     else if(!strncmp(smboxsep,buf,STRLEN(smboxsep))&&bogus)
 putsp:	lputcs(' ');
 #endif /* MAILBOX_SEPARATOR */
      lnl=buffilled==1;		      /* check if we just read an empty line */
@@ -798,7 +847,7 @@ putsp:	lputcs(' ');
 	       { lputssn(escap,escaplen);
 		 lputssn(chp,(p=strchr(chp,'\n')+1)-chp);
 	       }
-	      while((chp=p)<(char*)fldp->fld_text+fldp->tot_len);
+	      while((chp=p)<(char*)fldp->fld_text+fldp->Tot_len);
 	      free(fldp);					/* delete it */
 	    }
 	   while(fldp=fp2);		       /* escape all fields we found */
@@ -819,6 +868,7 @@ flbuf:	lputssn(buf,buffilled),buffilled=0;
   logfolder();
 onlyhead:
   closemine();
+nconlyhead:
   if(split)						/* wait for everyone */
    { int excode;
      close(STDIN);	       /* close stdin now, we're not reading anymore */
@@ -831,21 +881,29 @@ onlyhead:
   return retval!=EXIT_SUCCESS?retval:split<0?EX_IOERR:EXIT_SUCCESS;
 }
 
-eqFrom_(a)const char*const a;
+int eqFrom_(a)const char*const a;
 { return !strncmp(a,From_,STRLEN(From_));
 }
 
 int breakfield(line,len)const char*const line;size_t len;  /* look where the */
 { const char*p=line;			   /* fieldname ends (RFC 822 specs) */
-  if(eqFrom_(p))				      /* special case, From_ */
-     return STRLEN(From_);
-  while(len&&!iscntrl(*p))		    /* no control characters allowed */
+  while(len&&*p&&!iscntrl(*p))		    /* no control characters allowed */
    { switch(*p++)
       { default:len--;
 	   continue;
-	case HEAD_DELIMITER:len=p-line;
+	case HEAD_DELIMITER:
+good:	   len=p-line;
 	   return len==1?0:len;					  /* eureka! */
-	case ' ':p--;					/* no spaces allowed */
+	case ' ':case '\t':
+	   ;{ const char*q=p;	/* whitespace is okay right before the colon */
+	      while(--len&&(*q==' '||*q=='\t'))
+		 q++;
+	      if(len&&*q==HEAD_DELIMITER)
+		 goto good;
+	      if(eqFrom_(line))			      /* special case, From_ */
+		 return STRLEN(From_);
+	      p--;
+	    }					   /* it was bogus after all */
       }
      break;
    }
