@@ -1,12 +1,12 @@
 /************************************************************************
  *	Collection of standard library substitute routines		*
  *									*
- *	Copyright (c) 1990-1994, S.R. van den Berg, The Netherlands	*
+ *	Copyright (c) 1990-1997, S.R. van den Berg, The Netherlands	*
  *	#include "../README"						*
  ************************************************************************/
 #ifdef RCS
 static /*const*/char rcsid[]=
- "$Id: sublib.c,v 1.15 1994/10/07 15:25:11 berg Exp $";
+ "$Id: sublib.c,v 1.19 1999/04/19 06:42:28 guenther Exp $";
 #endif
 #include "includes.h"
 #include "sublib.h"
@@ -127,7 +127,7 @@ ret0:
 #endif
 			    /* strtol replacement which lacks range checking */
 #ifdef NOstrtol
-long strtol(start,ptr,base)const char*start,**const ptr;
+long strtol(start,ptr,base)const char*start,**const ptr;int base;
 { long result;const char*str=start;unsigned i;int sign,found;
   if(base>=36||base<(sign=found=result=0))
      goto fault;
